/*!
** @file
** @brief   The implementation of the main dialog box.
** @remarks This utility is a simple dialog-based application.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
** @license The MIT License
** @verbatim

Copyright (C) 2014 Ueta System Design Studio

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

@endverbatim
*/
#pragma once

#include "about.h"
#include "controls.h"
#include "dwmapix.h"
#include "enumerate.h"
#include "reticle.h"

/*!
** @brief   The main dialog box.
** @remarks -
*/
class CDialog :
		public CDialogImpl<CDialog>,
		public CWinDataExchange<CDialog>,
		public CAboutBoxSiteImpl<CDialog>,
		public CReticleSiteImpl<CDialog>,
		public CEnumDisplayMonitors,
		public CEnumTopLevelWindows,
		public CMessageFilter,
		public CIdleHandler
{
public:

	enum {IDD = IDR_APP}; // specifies the dialog template resource ID.

private:

	CAccelerator m_accel;

	CMonitorList m_wndMonitorList;
	CWindowList m_wndWindowList;
	CCoordValue m_wndCoordValues[4];

	HMONITOR m_hMonitor;
	MONITORINFO m_mi;
	HWND m_hWndTarget;
	WINDOWINFO m_wi;

	CRect m_rcDefault;
	CRect m_rcRestore;

public:

	/*!
	** @brief   Initializes a new CDialog object.
	** @remarks -
	*/
	CDialog()
	{
		m_hMonitor   = NULL;
		m_hWndTarget = NULL;

		::ZeroMemory(&m_mi, sizeof(m_mi));
		m_mi.cbSize = sizeof(MONITORINFO);
		::ZeroMemory(&m_wi, sizeof(m_wi));
		m_wi.cbSize = sizeof(WINDOWINFO);
	}

	/*!
	** @brief   Determines whether a window is suitable for the application.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns TRUE if suitable, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL IsWindowSuitable(HWND hWnd)
	{
		if (::IsWindowVisible(hWnd))
		{
			TITLEBARINFO ti = {sizeof(TITLEBARINFO)};

			if (::GetTitleBarInfo(hWnd, &ti))
			{
				if (!(ti.rgstate[0] & STATE_SYSTEM_INVISIBLE)) // the title bar itself
				{
					DWORD dwProcessId;
					DWORD dwThreadId = ::GetWindowThreadProcessId(hWnd, &dwProcessId);
					DBG_UNREFERENCED_LOCAL_VARIABLE(dwProcessId);

					if (dwThreadId)
					{
						if (dwThreadId != ::GetCurrentThreadId())
						{
							return TRUE;
						}
					}
				}
			}
		}
		return FALSE;
	}

	/*!
	** @brief   Retrieves the default dimensions of a window.
	** @param   [out] lpRect the bounding rectangle of a window, in screen coordinates.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL GetDefaultWindowRect(LPRECT lpRect)
	{
		static WINDOWINFO wi = {sizeof(WINDOWINFO)};

		if (::IsRectEmpty(&wi.rcWindow))
		{
			CStatic wndHidden;

			if (!wndHidden.Create(HWND_DESKTOP, NULL, NULL, WS_OVERLAPPEDWINDOW))
			{
				return FALSE;
			}
			ATLVERIFY(DwmGetWindowInfo(wndHidden, &wi));
			ATLASSERT(!::IsRectEmpty(&wi.rcWindow));
			ATLVERIFY(wndHidden.DestroyWindow());
		}
		ATLVERIFY(::CopyRect(lpRect, &wi.rcWindow));

		return TRUE;
	}

	/*!
	** @brief   Associates a large or small icon with the window.
	** @param   [in] icon specifies the icon resource to be loaded.
	** @param   [in] bBigIcon the type of the icon.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SetIcon(_U_STRINGorID icon, BOOL bBigIcon = TRUE)
	{
		int cxIcon = ::GetSystemMetrics(bBigIcon ? SM_CXICON : SM_CXSMICON);
		int cyIcon = ::GetSystemMetrics(bBigIcon ? SM_CYICON : SM_CYSMICON);

		if (HICON hIcon = AtlLoadIconImage(icon, LR_DEFAULTCOLOR, cxIcon, cyIcon))
		{
			CWindow::SetIcon(hIcon, bBigIcon);

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Enables or disables the specified control in the dialog box.
	** @param   [in] nCtrlID the identifier of the control to be enabled or disabled.
	** @param   [in] bEnable specifies whether the control is to be enabled or disabled.
	** @returns TRUE if the control was previously disabled, or FALSE otherwise.
	** @remarks -
	*/
	BOOL EnableDlgCtrl(int nCtrlID, BOOL bEnable = TRUE)
	{
		CWindow wndCtrl = GetDlgItem(nCtrlID);
		ATLASSERT(wndCtrl);

		return wndCtrl.EnableWindow(bEnable);
	}

	/*!
	** @brief   Enables or disables the interactive controls in the dialog box.
	** @param   [in] bEnable specifies whether the controls are to be enabled or disabled.
	** @returns No return value.
	** @remarks -
	*/
	void EnableDlgCtrls(BOOL bEnable = TRUE)
	{
		BOOL bThickFrame = m_wi.dwStyle & WS_THICKFRAME;

		BOOL bNormal = !(m_wi.dwStyle & (WS_MAXIMIZE | WS_MINIMIZE));

		EnableDlgCtrl(IDC_WIDTH,  bEnable && bNormal && bThickFrame);
		EnableDlgCtrl(IDC_HEIGHT, bEnable && bNormal && bThickFrame);
		EnableDlgCtrl(IDC_LEFT,   bEnable && bNormal);
		EnableDlgCtrl(IDC_TOP,    bEnable && bNormal);

		for (int iPad = 0; iPad < 9; ++iPad)
		{
			EnableDlgCtrl(IDC_PAD + iPad, bEnable && bNormal);
		}
		EnableDlgCtrl(IDC_APPLYNOW, bEnable && bNormal);

		if (!bEnable)
		{
			ATLVERIFY(SetDlgItemText(IDC_WIDTH , NULL));
			ATLVERIFY(SetDlgItemText(IDC_HEIGHT, NULL));
			ATLVERIFY(SetDlgItemText(IDC_LEFT,   NULL));
			ATLVERIFY(SetDlgItemText(IDC_TOP,    NULL));

			EnableDlgCtrl(IDC_RESTORE, FALSE);
		}
	}

	/*!
	** @brief   Moves the focus to the specified control in the dialog box.
	** @param   [in] nCtrlID the identifier of the control that is to receive the focus.
	** @returns No return value.
	** @remarks -
	*/
	void GotoDlgCtrl(int nCtrlID) const
	{
		CWindow wndCtrl = GetDlgItem(nCtrlID);
		ATLASSERT(wndCtrl);

		CWindow::GotoDlgCtrl(wndCtrl);
	}

	/*!
	** @copydoc CEnumDisplayMonitors::EnumDisplayMonitors
	*/
	BOOL EnumDisplayMonitors()
	{
		m_hMonitor = NULL;

		m_wndMonitorList.ResetContent();

		if (CEnumDisplayMonitors::EnumDisplayMonitors())
		{
			;

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @copydoc CEnumDisplayMonitors::EnumDisplayMonitorsProc
	*/
	virtual BOOL EnumDisplayMonitorsProc(HMONITOR hMonitor)
	{
		ATLASSERT(hMonitor);

		MONITORINFOEX mi;
		::ZeroMemory(&mi, sizeof(MONITORINFOEX));
		mi.cbSize = sizeof(MONITORINFOEX);
		ATLVERIFY(::GetMonitorInfo(hMonitor, &mi));

		DISPLAY_DEVICE display = {sizeof(DISPLAY_DEVICE)};

		if (::EnumDisplayDevices(mi.szDevice, 0, &display, 0))
		{
			;
		}
		else // Default Monitor
		{
			for (DWORD iDevNum = 0; ::EnumDisplayDevices(NULL, iDevNum, &display, 0); ++iDevNum)
			{
				if (lstrcmp(display.DeviceName, mi.szDevice) == 0)
				{
					break;
				}
			}
		}
		if (display.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP)
		{
			m_wndMonitorList.AddItem(hMonitor, display.DeviceString);
		}
		return TRUE;
	}

	/*!
	** @copydoc CEnumTopLevelWindows::EnumTopLevelWindows
	*/
	BOOL EnumTopLevelWindows()
	{
		EnableDlgCtrls(FALSE);

		m_hWndTarget = NULL;

		int iCurSel = m_wndMonitorList.GetCurSel();

		m_hMonitor = m_wndMonitorList.GetItem(iCurSel);
		ATLASSERT(m_hMonitor);
		ATLVERIFY(::GetMonitorInfo(m_hMonitor, &m_mi));
		ATLASSERT(!::IsRectEmpty(&m_mi.rcMonitor));
		ATLASSERT(!::IsRectEmpty(&m_mi.rcWork));

		m_wndWindowList.DeleteAllItems();

		if (CEnumTopLevelWindows::EnumTopLevelWindows())
		{
			ATLVERIFY(m_wndWindowList.SortItems());

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @copydoc CEnumTopLevelWindows::EnumTopLevelWindowsProc
	*/
	virtual BOOL EnumTopLevelWindowsProc(HWND hWnd)
	{
		if (IsWindowSuitable(hWnd))
		{
			HMONITOR hMonitor = ::MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
			ATLASSERT(hMonitor);
			DBG_UNREFERENCED_LOCAL_VARIABLE(hMonitor);

			WINDOWINFO wi = {sizeof(WINDOWINFO)};
			ATLVERIFY(DwmGetWindowInfo(hWnd, &wi));
			CRect rcWindow = wi.rcWindow;
			m_wndWindowList.AddItem(hWnd, rcWindow);
		}
		return TRUE;
	}

	/*!
	** @brief   Retrieves the current values from the coordinate value editors.
	** @param   [out] lpRect the bounding rectangle of a window, in screen coordinates.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL GetCoordValues(LPRECT lpRect)
	{
		CRect rcWindow = m_wi.rcWindow;

		int cx = rcWindow.Width();
		int cy = rcWindow.Height();
		int  x = rcWindow.left;
		int  y = rcWindow.top;

		int nVal;
		int nMin = CCoordValue::MIN;
		int nMax = CCoordValue::MAX;

		if (DDX_Int(IDC_WIDTH,  nVal, TRUE, DDX_SAVE, TRUE, nMin, nMax)) cx = nVal;
		if (DDX_Int(IDC_HEIGHT, nVal, TRUE, DDX_SAVE, TRUE, nMin, nMax)) cy = nVal;
		if (DDX_Int(IDC_LEFT,   nVal, TRUE, DDX_SAVE, TRUE, nMin, nMax))  x = nVal;
		if (DDX_Int(IDC_TOP,    nVal, TRUE, DDX_SAVE, TRUE, nMin, nMax))  y = nVal;

		ATLASSERT(lpRect);

		*lpRect = CRect(CPoint(x, y), CSize(cx, cy));

		return TRUE;
	}

	/*!
	** @brief   Exchanges and validates dialog data.
	** @param   [in] bSaveAndValidate indicates the direction of the data transfer.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL DoDataExchange(BOOL bSaveAndValidate = FALSE)
	{
		if (bSaveAndValidate)
		{
			CRect rcWindow;

			ATLVERIFY(GetCoordValues(rcWindow));

			rcWindow.NormalizeRect();
			m_wi.rcWindow = rcWindow;
		}
		else
		{
			CRect rcWindow = m_wi.rcWindow;

			int cx = rcWindow.Width();
			int cy = rcWindow.Height();
			int  x = rcWindow.left;
			int  y = rcWindow.top;

			ATLVERIFY(DDX_Int(IDC_WIDTH,  cx, TRUE, DDX_LOAD));
			ATLVERIFY(DDX_Int(IDC_HEIGHT, cy, TRUE, DDX_LOAD));
			ATLVERIFY(DDX_Int(IDC_LEFT,    x, TRUE, DDX_LOAD));
			ATLVERIFY(DDX_Int(IDC_TOP,     y, TRUE, DDX_LOAD));

			int iItem = m_wndWindowList.FindItem(m_hWndTarget);
			m_wndWindowList.SetSubItems(iItem, rcWindow);
			m_wndWindowList.SortItems(), iItem = -1;
		}
		return TRUE;
	}

	/*!
	** @brief   Called when a data exchange error has occurred.
	** @param   [in] nCtrlID the identifier of the control there is the error.
	** @param   [in] bSave indicates the direction of the data transfer.
	** @returns No return value.
	** @remarks -
	*/
	void OnDataExchangeError(UINT nCtrlID, BOOL bSave)
	{
		UNREFERENCED_PARAMETER(nCtrlID);
		UNREFERENCED_PARAMETER(bSave);

		;
	}

	/*!
	** @brief   Called when a data validation error has occurred.
	** @param   [in] nCtrlID the identifier of the control there is the error.
	** @param   [in] bSave indicates the direction of the data transfer.
	** @param   [in] data the data to report the details of the error.
	** @returns No return value.
	** @remarks -
	*/
	void OnDataValidateError(UINT nCtrlID, BOOL bSave, _XData& data)
	{
		UNREFERENCED_PARAMETER(nCtrlID);
		UNREFERENCED_PARAMETER(bSave);
		UNREFERENCED_PARAMETER(data);

		;
	}

	/*!
	** @brief   Applies the changes to the currently selected window.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL ApplyNow()
	{
		ATLASSERT(m_hWndTarget);

		DWORD dwStyle = DWORD(::GetWindowLong(m_hWndTarget, GWL_STYLE));

		if (!(dwStyle & (WS_MAXIMIZE | WS_MINIMIZE)))
		{
			CRect rcWindow = m_wi.rcWindow;

			int cx = rcWindow.Width();
			int cy = rcWindow.Height();
			int  x = rcWindow.left;
			int  y = rcWindow.top;

			if (DwmSetWindowPos(m_hWndTarget, HWND_TOP, x, y, cx, cy, 0))
			{
				DwmGetWindowInfo(m_hWndTarget, &m_wi); // continue to run even if failed
				ATLVERIFY(BringWindowToTop());
				ATLVERIFY(DoDataExchange(DDX_LOAD));

				return TRUE;
			}
			else // invalid handle, probably already destroyed
			{
				ATLVERIFY(EnumTopLevelWindows());

				ATLVERIFY(::MessageBeep(MB_ICONERROR));

				return FALSE;
			}
		}
		else // minimized window or maximized window
		{
			if (DwmGetWindowInfo(m_hWndTarget, &m_wi))
			{
				m_rcRestore = m_wi.rcWindow;
				ATLVERIFY(DoDataExchange(DDX_LOAD));
				EnableDlgCtrls();

				ATLVERIFY(::MessageBeep(MB_ICONERROR));

				return FALSE;
			}
			else // invalid handle, probably already destroyed
			{
				ATLVERIFY(EnumTopLevelWindows());

				ATLVERIFY(::MessageBeep(MB_ICONERROR));

				return FALSE;
			}
		}
	}

	BEGIN_MSG_MAP(CDialog)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_DESTROY(OnDestroy)
		MSG_WM_CONTEXTMENU(OnContextMenu)
		COMMAND_ID_HANDLER_EX(IDCANCEL, OnCancel)
		COMMAND_ID_HANDLER_EX(IDC_APPLYNOW, OnApplyNow)
		COMMAND_ID_HANDLER_EX(IDC_RESTORE, OnRestore)
		COMMAND_ID_HANDLER_EX(IDM_REFRESH, OnRefresh)
		COMMAND_HANDLER_EX(IDC_MONITORLIST, CBN_SELENDOK, OnMonitorListSelEndOk)
		NOTIFY_HANDLER_EX(IDC_WINDOWLIST, LVN_ITEMCHANGED, OnWindowListItemChanged)
		COMMAND_CODE_HANDLER_EX(BN_CLICKED, OnClicked)
		CHAIN_MSG_MAP(CAboutBoxSiteImpl<CDialog>)
		REFLECT_NOTIFICATIONS()
	END_MSG_MAP()

	/*!
	** @brief   Filters window messages before they are dispatched to windows.
	** @param   [in] pMsg a pointer to an MSG structure that contains a message.
	** @returns TRUE if the message should not be dispatched, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL PreTranslateMessage(MSG* pMsg)
	{
		if (!m_accel.TranslateAccelerator(m_hWnd, pMsg))
		{
			if  (!IsDialogMessage(pMsg))
			{
				return FALSE;
			}
		}
		return TRUE;
	}

	/*!
	** @brief   Called when no messages are pending on the message queue.
	** @returns TRUE to do further processing, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL OnIdle()
	{
		;

		return FALSE;
	}

	/*!
	** @brief   Called in response to the WM_INITDIALOG message.
	** @param   [in] hWndFocus a handle to the control to receive the default input focus.
	** @param   [in] lInitParam additional initialization data.
	** @returns TRUE to set the default input focus, or FALSE otherwise.
	** @remarks -
	*/
	BOOL OnInitDialog(HWND hWndFocus, LPARAM lInitParam)
	{
		UNREFERENCED_PARAMETER(hWndFocus);
		UNREFERENCED_PARAMETER(lInitParam);

		ATLVERIFY(ExecuteDlgInit(IDD));

		ATLVERIFY(SetIcon(IDR_APP, TRUE));
		ATLVERIFY(SetIcon(IDR_APP, FALSE));

		ATLVERIFY(GetDefaultWindowRect(m_rcDefault));

		DDX_Control(IDC_MONITORLIST, m_wndMonitorList, DDX_LOAD);
		DDX_Control(IDC_WINDOWLIST,  m_wndWindowList,  DDX_LOAD);

		DDX_Control(IDC_WIDTH,  m_wndCoordValues[0], DDX_LOAD);
		DDX_Control(IDC_HEIGHT, m_wndCoordValues[1], DDX_LOAD);
		DDX_Control(IDC_LEFT,   m_wndCoordValues[2], DDX_LOAD);
		DDX_Control(IDC_TOP,    m_wndCoordValues[3], DDX_LOAD);

		ATLVERIFY(m_accel.LoadAccelerators(IDR_APP));

		ATLVERIFY(BindAboutBox());
		ATLVERIFY(BindReticle());

		ATLVERIFY(EnumDisplayMonitors());
		m_wndMonitorList.SetCurSel(0);
		ATLVERIFY(EnumTopLevelWindows());

		ATLVERIFY(CenterWindow());

		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->AddMessageFilter(this));
		ATLVERIFY(pMessageLoop->AddIdleHandler(this));

		return TRUE;
	}

	/*!
	** @brief   Called when the dialog box is about to be destroyed.
	** @returns No return value.
	** @remarks -
	*/
	void OnDestroy()
	{
		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->RemoveMessageFilter(this));
		ATLVERIFY(pMessageLoop->RemoveIdleHandler(this));

		::PostQuitMessage(int(S_OK));
	}

	/*!
	** @brief   Called when the user has clicked the right mouse button.
	** @param   [in] hWnd a handle to the window in which the user right-clicked the mouse.
	** @param   [in] point the position of the cursor, in screen coordinates.
	** @returns No return value.
	** @remarks -
	*/
	void OnContextMenu(HWND hWnd, CPoint point)
	{
		CMenu menu;
		ATLVERIFY(menu.LoadMenu(IDR_APP));
		ATLASSERT(menu.GetMenuItemCount() == 1);

		switch (::GetDlgCtrlID(hWnd))
		{
		case IDC_WINDOWLIST:
			ATLVERIFY(menu.GetSubMenu(0).TrackPopupMenu(0, point.x, point.y, m_hWnd));
			break;

		default:
			break;
		}
	}

	/*!
	** @brief   Called when the user has clicked the Close button or has pressed the ESC key.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnCancel(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		DestroyWindow();
	}

	/*!
	** @brief   Called when the user has clicked the Apply button.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnApplyNow(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		ATLVERIFY(DoDataExchange(DDX_SAVE));

		if (ApplyNow() && (m_rcRestore != m_wi.rcWindow))
		{
			EnableDlgCtrl(IDC_RESTORE, TRUE);
		}
	}

	/*!
	** @brief   Called when the user has clicked the Restore button.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnRestore(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		m_wi.rcWindow = m_rcRestore;

		if (ApplyNow() && (m_rcRestore == m_wi.rcWindow))
		{
			EnableDlgCtrl(IDC_RESTORE, FALSE);
		}
	}

	/*!
	** @brief   Called when the user executes the Refresh command.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnRefresh(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		ATLVERIFY(EnumTopLevelWindows());
	}

	/*!
	** @copydoc CReticleSite::OnReticleEnter
	*/
	virtual BOOL OnReticleEnter(HWND hWnd)
	{
		if (IsWindowSuitable(hWnd))
		{
			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @copydoc CReticleSite::OnReticleRelease
	*/
	virtual void OnReticleRelease(HWND hWnd)
	{
		if (IsWindowSuitable(hWnd))
		{
			HMONITOR hMonitor = ::MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
			ATLASSERT(hMonitor);
			ATLVERIFY(EnumDisplayMonitors());
			if (m_wndMonitorList.SetCurSel(hMonitor) == CB_ERR)
				m_wndMonitorList.SetCurSel(0);
			ATLVERIFY(EnumTopLevelWindows());
			m_wndWindowList.SelectItem(hWnd);

			GotoDlgCtrl(IDC_WINDOWLIST);
		}
	}

	/*!
	** @brief   Handles the CBN_SELENDOK notification from the CMonitorList object.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @returns No return value.
	** @remarks -
	*/
	void OnMonitorListSelEndOk(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iCurSel = m_wndMonitorList.GetCurSel();

		HMONITOR hMonitor = m_wndMonitorList.GetItem(iCurSel);
		ATLASSERT(hMonitor);
		ATLVERIFY(EnumDisplayMonitors());
		if (m_wndMonitorList.SetCurSel(hMonitor) == CB_ERR)
			m_wndMonitorList.SetCurSel(0);
		ATLVERIFY(EnumTopLevelWindows());
	}

	/*!
	** @brief   Handles the LVN_ITEMCHANGED notification from the CWindowList object.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnWindowListItemChanged(LPNMHDR pnmh)
	{
		LPNMLISTVIEW lpNMListView = reinterpret_cast<LPNMLISTVIEW>(pnmh);

		ATLASSERT(lpNMListView);

		if (lpNMListView->uChanged & LVIF_STATE)
		{
			UINT uNewState = lpNMListView->uNewState;
			UINT uOldState = lpNMListView->uOldState;

			UINT stateMask = uNewState ^ uOldState;

			if (stateMask & LVIS_SELECTED)
			{
				if (uNewState & LVIS_SELECTED)
				{
					ATLASSERT(!m_hWndTarget);
					m_hWndTarget = HWND(lpNMListView->lParam);
					ATLASSERT(m_hWndTarget);

					if (DwmGetWindowInfo(m_hWndTarget, &m_wi))
					{
						m_rcRestore = m_wi.rcWindow;
						ATLVERIFY(DoDataExchange(DDX_LOAD));
						EnableDlgCtrls();
					}
					else // invalid handle, probably already destroyed
					{
						ATLVERIFY(EnumTopLevelWindows());
					}
				}
				else // deselect
				{
					ATLASSERT(m_hWndTarget);
					EnableDlgCtrls(FALSE);
					m_hWndTarget = NULL;
				}
			}
		}
		return LRESULT(0); // void
	}

	/*!
	** @brief   Handles the BN_CLICKED notification from any button or menu item.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the menu item, control, or accelerator.
	** @param   [in] hWndCtrl a handle to the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnClicked(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(hWndCtrl);

		CRect rcPreserve = m_wi.rcWindow;
		CRect rcFull     = m_mi.rcWork;
		CRect rcWindow;

		ATLVERIFY(GetCoordValues(rcWindow));

		CRect rcOrigin = rcWindow + (rcFull.TopLeft()     - rcWindow.TopLeft());
		CRect rcCenter = rcWindow + (rcFull.CenterPoint() - rcWindow.CenterPoint());
		CRect rcBottom = rcWindow + (rcFull.BottomRight() - rcWindow.BottomRight());

		switch (nID)
		{
		case IDM_WIDTH_PRESERVE:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH, rcPreserve.Width()));
			GotoDlgCtrl(IDC_WIDTH);
			break;
		case IDM_WIDTH_DEFAULT:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH, m_rcDefault.Width()));
			GotoDlgCtrl(IDC_WIDTH);
			break;
		case IDM_WIDTH_FULL:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH, rcFull.Width()));
			GotoDlgCtrl(IDC_WIDTH);
			break;
		case IDM_WIDTH_HALF:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH, rcFull.Width() / 2));
			GotoDlgCtrl(IDC_WIDTH);
			break;
		case IDM_WIDTH_QUARTER:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH, rcFull.Width() / 4));
			GotoDlgCtrl(IDC_WIDTH);
			break;

		case IDM_HEIGHT_PRESERVE:
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcPreserve.Height()));
			GotoDlgCtrl(IDC_HEIGHT);
			break;
		case IDM_HEIGHT_DEFAULT:
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, m_rcDefault.Height()));
			GotoDlgCtrl(IDC_HEIGHT);
			break;
		case IDM_HEIGHT_FULL:
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcFull.Height()));
			GotoDlgCtrl(IDC_HEIGHT);
			break;
		case IDM_HEIGHT_HALF:
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcFull.Height() / 2));
			GotoDlgCtrl(IDC_HEIGHT);
			break;
		case IDM_HEIGHT_QUARTER:
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcFull.Height() / 4));
			GotoDlgCtrl(IDC_HEIGHT);
			break;

		case IDM_LEFT_PRESERVE:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcPreserve.left));
			GotoDlgCtrl(IDC_LEFT);
			break;
		case IDM_LEFT_DEFAULT:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   m_rcDefault.left));
			GotoDlgCtrl(IDC_LEFT);
			break;
		case IDM_LEFT_LEFTALIGN:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcOrigin.left));
			GotoDlgCtrl(IDC_LEFT);
			break;
		case IDM_LEFT_CENTER:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcCenter.left));
			GotoDlgCtrl(IDC_LEFT);
			break;
		case IDM_LEFT_RIGHTALIGN:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcBottom.left));
			GotoDlgCtrl(IDC_LEFT);
			break;

		case IDM_TOP_PRESERVE:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcPreserve.top));
			GotoDlgCtrl(IDC_TOP);
			break;
		case IDM_TOP_DEFAULT:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    m_rcDefault.top));
			GotoDlgCtrl(IDC_TOP);
			break;
		case IDM_TOP_TOPALIGN:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcOrigin.top));
			GotoDlgCtrl(IDC_TOP);
			break;
		case IDM_TOP_CENTER:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcCenter.top));
			GotoDlgCtrl(IDC_TOP);
			break;
		case IDM_TOP_BOTTOMALIGN:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcBottom.top));
			GotoDlgCtrl(IDC_TOP);
			break;

		case IDC_PAD_CENTER:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcCenter.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcCenter.top));
			break;
		case IDC_PAD_TOP:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcCenter.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcOrigin.top));
			break;
		case IDC_PAD_BOTTOM:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcCenter.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcBottom.top));
			break;
		case IDC_PAD_LEFT:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcOrigin.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcCenter.top));
			break;
		case IDC_PAD_LEFTTOP:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcOrigin.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcOrigin.top));
			break;
		case IDC_PAD_LEFTBOTTOM:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcOrigin.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcBottom.top));
			break;
		case IDC_PAD_RIGHT:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcBottom.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcCenter.top));
			break;
		case IDC_PAD_RIGHTTOP:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcBottom.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcOrigin.top));
			break;
		case IDC_PAD_RIGHTBOTTOM:
			ATLVERIFY(SetDlgItemInt(IDC_WIDTH,  rcWindow.Width()));
			ATLVERIFY(SetDlgItemInt(IDC_HEIGHT, rcWindow.Height()));
			ATLVERIFY(SetDlgItemInt(IDC_LEFT,   rcBottom.left));
			ATLVERIFY(SetDlgItemInt(IDC_TOP,    rcBottom.top));
			break;

		default:
			SetMsgHandled(FALSE);
			return;
		}
	}
};
