/*!
** @file
** @brief   The entry point of the program.
** @remarks This file is the single compilation unit of the program.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
** @license The MIT License
** @verbatim

Copyright (C) 2014 Ueta System Design Studio

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

@endverbatim
*/
#include "stdafx.h"
#include "resource.h"
#include "dialog.h"

/*!
** @brief   The global instance of CAppModule.
** @remarks -
*/
CAppModule _Module;

/*!
** @brief   The dialog box manipulator.
** @remarks -
*/
class CManipulator : public CWindow
{
public:

	/*!
	** @brief   Constructs a new CManipulator object to manipulate the specified dialog box.
	** @param   [in] hDlg a handle to the dialog box to be manipulated.
	** @remarks -
	*/
	CManipulator(HWND hDlg = NULL) : CWindow(hDlg)
	{
		;
	}

	/*!
	** @brief   Simulates clicking a Button control in the dialog box.
	** @param   [in] nCtrlID the identifier of the control.
	** @returns No return value.
	** @remarks -
	*/
	void Click(int nCtrlID)
	{
		CButton wndButton = GetDlgItem(nCtrlID);

		GotoDlgCtrl(wndButton);

		wndButton.Click();
	}

	/*!
	** @brief   Translates the text of a specified control in a dialog box into an integer value.
	** @param   [in] nCtrlID the identifier of the control.
	** @returns The translated value of the control text if successful, or zero otherwise.
	** @remarks -
	*/
	int GetDlgItemInt(int nCtrlID)
	{
		return int(CWindow::GetDlgItemInt(nCtrlID));
	}

	/*!
	** @brief   Pauses the execution of the current thread for the specified time.
	** @param   [in] dwMilliseconds the time, in milliseconds, for which to suspend execution.
	** @returns No return value.
	** @remarks -
	*/
	static void Pause(DWORD dwMilliseconds = 500)
	{
		UINT_PTR uIDEvent = ::SetTimer(NULL, 0, dwMilliseconds, NULL);
		ATLASSERT(uIDEvent);

		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		MSG& msg = pMessageLoop->m_msg;

		for (bool bDoIdle = true; ; )
		{
			if (bDoIdle && !::PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
			{
				pMessageLoop->OnIdle(0);

				bDoIdle = false;
			}
			if (!::GetMessage(&msg, NULL, 0, 0))
			{
				break;
			}
			if ((msg.message == WM_TIMER) && (msg.hwnd == NULL) && (msg.wParam == WPARAM(uIDEvent)))
			{
				break;
			}
			if (!pMessageLoop->PreTranslateMessage(&msg))
			{
				::TranslateMessage(&msg);
				::DispatchMessage(&msg);
			}
			if (CMessageLoop::IsIdleMessage(&msg))
			{
				bDoIdle = true;
			}
		}
		ATLVERIFY(::KillTimer(NULL, uIDEvent));
	}

	/*!
	** @brief   Selects the item in the List-View control that has the specified handle.
	** @param   [in] hWnd a handle to the window.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SelectItem(HWND hWnd)
	{
		CListViewCtrl wndListViewCtrl = GetDlgItem(IDC_WINDOWLIST);

		if (!hWnd)
		{
			return wndListViewCtrl.SetItemState(-1, 0, LVIS_SELECTED);
		}
		LVFINDINFO lvfi = {LVFI_PARAM};
		lvfi.lParam = LPARAM(hWnd);
		int iItem = wndListViewCtrl.FindItem(&lvfi);
		return wndListViewCtrl.SelectItem(iItem);
	}

	/*!
	** @brief   Sets the currently selected item in a Combo Box control.
	** @param   [in] nCtrlID the identifier of the control.
	** @param   [in] nSelect the zero-based index of the item to select.
	** @returns the index of the item selected if successful, or CB_ERR otherwise.
	** @remarks -
	*/
	int SetCurSel(int nCtrlID, int nSelect)
	{
		CComboBox wndComboBox = GetDlgItem(nCtrlID);

		int nSelected = wndComboBox.SetCurSel(nSelect);

		if (nSelected != CB_ERR)
		{
			SendMessage(WM_COMMAND, MAKEWPARAM(nCtrlID, CBN_SELENDOK), LPARAM(HWND(wndComboBox)));
		}
		return nSelected;
	}
};

/*!
** @brief   The dummy window.
** @remarks -
*/
class CDummy : public CStatic
{
public:

};

static CDummy wndDummy;

static CHandle hThread;
static CHandle hThreadEvent;

static unsigned __stdcall thread(void* arglist)
{
	LPRECT lpRect = reinterpret_cast<LPRECT>(arglist);

	CMessageLoop theMessageLoop;
	ATLVERIFY(_Module.AddMessageLoop(&theMessageLoop));
	ATLVERIFY(wndDummy.Create(HWND_DESKTOP, lpRect, NULL, WS_OVERLAPPEDWINDOW | WS_VISIBLE));
	ATLVERIFY(::SetEvent(hThreadEvent));
	unsigned retval = theMessageLoop.Run();
	ATLVERIFY(wndDummy.DestroyWindow());
	ATLVERIFY(_Module.RemoveMessageLoop());

	return retval;
}

/*!
** @brief   Implements the entry point for the application.
** @param   [in] hInstance a handle to the current instance of the application.
** @param   [in] hPrevInstance a legacy from 16-bit Windows, now always zero.
** @param   [in] lpCmdLine the command-line for the application.
** @param   [in] nCmdShow specifies how the window is to be shown.
** @returns A standard HRESULT value as the exit value.
** @remarks -
*/
extern "C" int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	UNREFERENCED_PARAMETER(hInstance);
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(nCmdShow);

	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	HRESULT hResult = S_OK;

	hResult = ::CoInitialize(NULL);
	ATLASSERT(SUCCEEDED(hResult));
	::DefWindowProc(NULL, WM_NULL, 0, 0L);
	AtlInitCommonControls(ICC_WIN95_CLASSES);
	hResult = _Module.Init(NULL, hInstance);
	ATLASSERT(SUCCEEDED(hResult));

	#if 1 // Setup
	{
		CRect rcDummy(CPoint(13, 13), CSize(641, 479));

		hThreadEvent.Attach(::CreateEvent(NULL, FALSE, FALSE, NULL));
		hThread.Attach(HANDLE(_beginthreadex(NULL, 0, thread, rcDummy, 0, NULL)));
		ATLVERIFY(::WaitForSingleObject(hThreadEvent, INFINITE) == WAIT_OBJECT_0);
		ATLASSERT(wndDummy.IsWindow());
	}
	#endif

	CMessageLoop theMessageLoop;

	ATLVERIFY(_Module.AddMessageLoop(&theMessageLoop));

	CDialog theDialog;
	ATLVERIFY(theDialog.Create(HWND_DESKTOP));
	ATLASSERT(theDialog.IsWindowVisible());
	CManipulator manipulator = theDialog;

	CRect rcWorkArea;
	ATLVERIFY(::SystemParametersInfo(SPI_GETWORKAREA, 0, rcWorkArea, 0));
	CRect rcDefault;
	ATLVERIFY(CDialog::GetDefaultWindowRect(rcDefault));
	CRect rcDummy;
	ATLVERIFY(wndDummy.GetWindowRect(rcDummy));

	CManipulator::Pause(1000);

	#if 1 // Test 1: Sizing
	{
		ATLVERIFY(manipulator.SelectItem(wndDummy));
		CManipulator::Pause();

		#if 1 // Width
		{
			ATLVERIFY(manipulator.SetCurSel(IDC_WIDTH, 4) != CB_ERR); // Quarter
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH) == rcWorkArea.Width() / 4);

			ATLVERIFY(manipulator.SetCurSel(IDC_WIDTH, 3) != CB_ERR); // Half
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH) == rcWorkArea.Width() / 2);

			ATLVERIFY(manipulator.SetCurSel(IDC_WIDTH, 2) != CB_ERR); // Full
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH) == rcWorkArea.Width());

			ATLVERIFY(manipulator.SetCurSel(IDC_WIDTH, 1) != CB_ERR); // Default
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH) == rcDefault.Width());

			ATLVERIFY(manipulator.SetCurSel(IDC_WIDTH, 0) != CB_ERR); // Preserve
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH) == rcDummy.Width());
		}
		#endif

		#if 1 // Height
		{
			ATLVERIFY(manipulator.SetCurSel(IDC_HEIGHT, 4) != CB_ERR); // Quarter
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcWorkArea.Height() / 4);

			ATLVERIFY(manipulator.SetCurSel(IDC_HEIGHT, 3) != CB_ERR); // Half
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcWorkArea.Height() / 2);

			ATLVERIFY(manipulator.SetCurSel(IDC_HEIGHT, 2) != CB_ERR); // Full
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcWorkArea.Height());

			ATLVERIFY(manipulator.SetCurSel(IDC_HEIGHT, 1) != CB_ERR); // Default
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcDefault.Height());

			ATLVERIFY(manipulator.SetCurSel(IDC_HEIGHT, 0) != CB_ERR); // Preserve
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcDummy.Height());
		}
		#endif

		#if 1 // Apply and Restore
		{
			CRect rcExpected(rcDummy.TopLeft(), rcDefault.Size());

			ATLVERIFY(manipulator.SetDlgItemInt(IDC_WIDTH,  rcExpected.Width()));
			ATLVERIFY(manipulator.SetDlgItemInt(IDC_HEIGHT, rcExpected.Height()));

			manipulator.Click(IDC_APPLYNOW);
			CManipulator::Pause();
			CRect rcApplied;
			ATLVERIFY(wndDummy.GetWindowRect(rcApplied));
			ATLASSERT(rcApplied == rcExpected);

			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH)  == rcExpected.Width());
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcExpected.Height());

			manipulator.Click(IDC_RESTORE);
			CManipulator::Pause();
			CRect rcRestored;
			ATLVERIFY(wndDummy.GetWindowRect(rcRestored));
			ATLASSERT(rcRestored == rcDummy);

			ATLASSERT(manipulator.GetDlgItemInt(IDC_WIDTH)  == rcDummy.Width());
			ATLASSERT(manipulator.GetDlgItemInt(IDC_HEIGHT) == rcDummy.Height());
		}
		#endif

		ATLVERIFY(manipulator.SelectItem(NULL));
		CManipulator::Pause();
	}
	#endif

	#if 1 // Test 2: Positioning
	{
		ATLVERIFY(manipulator.SelectItem(wndDummy));
		CManipulator::Pause();

		CRect rcOrigin = rcDummy + (rcWorkArea.TopLeft()     - rcDummy.TopLeft());
		CRect rcCenter = rcDummy + (rcWorkArea.CenterPoint() - rcDummy.CenterPoint());
		CRect rcBottom = rcDummy + (rcWorkArea.BottomRight() - rcDummy.BottomRight());

		#if 1 // Left
		{
			ATLVERIFY(manipulator.SetCurSel(IDC_LEFT, 4) != CB_ERR); // Right
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcBottom.left);

			ATLVERIFY(manipulator.SetCurSel(IDC_LEFT, 3) != CB_ERR); // Center
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcCenter.left);

			ATLVERIFY(manipulator.SetCurSel(IDC_LEFT, 2) != CB_ERR); // Left
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcOrigin.left);

			ATLVERIFY(manipulator.SetCurSel(IDC_LEFT, 1) != CB_ERR); // Default
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcDefault.left);

			ATLVERIFY(manipulator.SetCurSel(IDC_LEFT, 0) != CB_ERR); // Preserve
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcDummy.left);
		}
		#endif

		#if 1 // Top
		{
			ATLVERIFY(manipulator.SetCurSel(IDC_TOP, 4) != CB_ERR); // Bottom
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP) == rcBottom.top);

			ATLVERIFY(manipulator.SetCurSel(IDC_TOP, 3) != CB_ERR); // Center
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP) == rcCenter.top);

			ATLVERIFY(manipulator.SetCurSel(IDC_TOP, 2) != CB_ERR); // Top
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP) == rcOrigin.top);

			ATLVERIFY(manipulator.SetCurSel(IDC_TOP, 1) != CB_ERR); // Default
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP) == rcDefault.top);

			ATLVERIFY(manipulator.SetCurSel(IDC_TOP, 0) != CB_ERR); // Preserve
			CManipulator::Pause();
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP) == rcDummy.top);
		}
		#endif

		#if 1 // Pads, Apply, and Restore
		{
			for (int nCtrlID = IDC_PAD; nCtrlID < IDC_PAD + 9; ++nCtrlID)
			{
				CRect rcExpected;

				switch (nCtrlID)
				{
				case IDC_PAD_CENTER:
					rcExpected = CRect(CPoint(rcCenter.left, rcCenter.top), rcDummy.Size()); break;
				case IDC_PAD_TOP:
					rcExpected = CRect(CPoint(rcCenter.left, rcOrigin.top), rcDummy.Size()); break;
				case IDC_PAD_BOTTOM:
					rcExpected = CRect(CPoint(rcCenter.left, rcBottom.top), rcDummy.Size()); break;
				case IDC_PAD_LEFT:
					rcExpected = CRect(CPoint(rcOrigin.left, rcCenter.top), rcDummy.Size()); break;
				case IDC_PAD_LEFTTOP:
					rcExpected = CRect(CPoint(rcOrigin.left, rcOrigin.top), rcDummy.Size()); break;
				case IDC_PAD_LEFTBOTTOM:
					rcExpected = CRect(CPoint(rcOrigin.left, rcBottom.top), rcDummy.Size()); break;
				case IDC_PAD_RIGHT:
					rcExpected = CRect(CPoint(rcBottom.left, rcCenter.top), rcDummy.Size()); break;
				case IDC_PAD_RIGHTTOP:
					rcExpected = CRect(CPoint(rcBottom.left, rcOrigin.top), rcDummy.Size()); break;
				case IDC_PAD_RIGHTBOTTOM:
					rcExpected = CRect(CPoint(rcBottom.left, rcBottom.top), rcDummy.Size()); break;
				}
				manipulator.Click(nCtrlID);
				CManipulator::Pause(200);
				manipulator.Click(IDC_APPLYNOW);
				CManipulator::Pause();
				CRect rcApplied;
				ATLVERIFY(wndDummy.GetWindowRect(rcApplied));
				ATLASSERT(rcApplied == rcExpected);

				ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcExpected.left);
				ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP)  == rcExpected.top);
			}
			manipulator.Click(IDC_RESTORE);
			CManipulator::Pause();
			CRect rcRestored;
			ATLVERIFY(wndDummy.GetWindowRect(rcRestored));
			ATLASSERT(rcRestored == rcDummy);

			ATLASSERT(manipulator.GetDlgItemInt(IDC_LEFT) == rcDummy.left);
			ATLASSERT(manipulator.GetDlgItemInt(IDC_TOP)  == rcDummy.top);
		}
		#endif

		ATLVERIFY(manipulator.SelectItem(NULL));
		CManipulator::Pause();
	}
	#endif

	ATLVERIFY(theDialog.DestroyWindow());

	ATLVERIFY(_Module.RemoveMessageLoop());

	#if 1 // Teardown
	{
		DWORD idThread = ::GetThreadId(hThread);
		ATLVERIFY(::PostThreadMessage(idThread, WM_QUIT, WPARAM(), LPARAM()));
		ATLVERIFY(::WaitForSingleObject(hThread, INFINITE) == WAIT_OBJECT_0);
	}
	#endif

	_Module.Term();
	::CoUninitialize();
	return int(hResult);
}
